/* TinyBasic (Yacc/Bison parser definition file.)
 * Version 0.1 (13/09-08)
 * Copyright 2008 Kjetil Erga (kobolt.anarion -AT- gmail -DOT- com)
 *
 *
 * This file is part of TinyBasic.
 *
 * TinyBasic is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * TinyBasic is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with TinyBasic. If not, see <http://www.gnu.org/licenses/>.
 */

%{
#include "tb-ast.h"
#include "tb-prg.h"
#include "y.tab.h"
#include <stdlib.h> /* For NULL definition. */
#include <stdio.h>

static int parse_line = -1; /* To locate syntax errors more easily. */

void yyerror(char *s);
int yylex(YYSTYPE *lvalp);
%}

%pure-parser
%error-verbose

%union {
  void *generic;
  char var;
  int number;
  char *string;
}

%token PRINT
%token IF
%token THEN
%token GOTO
%token INPUT
%token LET
%token GOSUB
%token RETURN
%token CLEAR
%token LIST
%token RUN
%token END
%token RND
%token USR
%token LOAD
%token SAVE
%token TREELIST
%token SYSTEM

%left '+' '-'
%left '*' '/'
%left EQ NE GE LE GT LT
%left UPLUS UMINUS

%token COMMA
%token NEWLINE

%token <var> VAR
%token <number> NUMBER
%token <string> STRING REM

%type <generic> remark statement expr_list var_list expression relop
  
%%

program : line
        | program line
        ;

line : NEWLINE                    /* Empty line, ignore. */
     | REM NEWLINE              { free(yylval.string); } /* Ignore, and free. */
     | NUMBER NEWLINE           { prg_add_line($1, NULL); }
     | NUMBER statement NEWLINE { parse_line = $1; prg_add_line($1, $2); }
     | statement NEWLINE        { parse_line = -1; ast_exec($1); ast_free($1); }
     | NUMBER remark NEWLINE    { parse_line = $1; prg_add_line($1, $2); }
     ;

/* Note: BASIC Remarks must be part of the program to be shown in listings. */
remark : REM { $$ = ast_add_node_string(AST_REM, NULL, $1);
               free(yylval.string); }
       ;

statement : PRINT expr_list       { $$ = ast_add_node_sub(AST_PRINT, $2); }
          | IF expression relop expression THEN statement
            { $$ = ast_add_node_complex(AST_IF, $6, $2, $3, $4); }
          | GOTO expression       { $$ = ast_add_node_sub(AST_GOTO, $2); }
          | INPUT var_list        { $$ = ast_add_node_sub(AST_INPUT, $2); }
          | LET VAR EQ expression { $$ = ast_add_node_var(AST_LET, $4, $2); }
          | GOSUB expression      { $$ = ast_add_node_sub(AST_GOSUB, $2); }
          | RETURN                { $$ = ast_add_node_leaf(AST_RETURN); }
          | CLEAR                 { $$ = ast_add_node_leaf(AST_CLEAR); }
          | LIST                  { $$ = ast_add_node_leaf(AST_LIST); }
          | RUN                   { $$ = ast_add_node_leaf(AST_RUN); }
          | END                   { $$ = ast_add_node_leaf(AST_END); }
          | LOAD expr_list        { $$ = ast_add_node_sub(AST_LOAD, $2); }
          | SAVE expr_list        { $$ = ast_add_node_sub(AST_SAVE, $2); }
          | TREELIST              { $$ = ast_add_node_leaf(AST_TREELIST); }
          ;

expr_list : { $$ = NULL; } /* Empty expression list. */
          | STRING     
            { $$ = ast_add_node_string(AST_STRING, NULL, $1);
              free(yylval.string); }
          | expression
            { $$ = ast_add_node_expr(AST_EXPR, NULL, $1); }
          | expr_list COMMA STRING
            { $$ = ast_add_node_string(AST_STRING, $1, $3);
              free(yylval.string); }
          | expr_list COMMA expression
            { $$ = ast_add_node_expr(AST_EXPR, $1, $3); }
          ;

var_list : VAR                { $$ = ast_add_node_var(AST_VAR, NULL, $1); }
         | var_list COMMA VAR { $$ = ast_add_node_var(AST_VAR, $1, $3); }
         ;

/* NOTE: 'term' and 'factor' removed for simpler yacc rules. */
expression : VAR   
             { $$ = ast_add_node_var(AST_VAR, NULL, $1); }
           | NUMBER
             { $$ = ast_add_node_number(AST_NUMBER, $1); }
           | expression '+' expression
             { $$ = ast_add_node_expr(AST_PLUS, $1, $3); }
           | expression '-' expression
             { $$ = ast_add_node_expr(AST_MINUS, $1, $3); }
           | expression '*' expression
             { $$ = ast_add_node_expr(AST_MULTIPLY, $1, $3); }
           | expression '/' expression
             { $$ = ast_add_node_expr(AST_DIVIDE, $1, $3); }
           | '+' expression %prec UPLUS
             { $$ = ast_add_node_expr(AST_PLUS, NULL, $2); }
           | '-' expression %prec UMINUS
             { $$ = ast_add_node_expr(AST_MINUS, NULL, $2); }
           | '(' expression ')'
             { $$ = $2; }
           | RND '(' expression ')'
             { $$ = ast_add_node_expr(AST_RND, NULL, $3); }
           | USR '(' expression ')'
             { $$ = ast_add_node_expr(AST_USR, NULL, $3); }
           | USR '(' expression COMMA expression ')'
             { $$ = ast_add_node_complex(AST_USR, NULL, $3, $5, NULL); }
           | USR '(' expression COMMA expression COMMA expression ')'
             { $$ = ast_add_node_complex(AST_USR, NULL, $3, $5, $7); }
           | SYSTEM '(' STRING ')'
             { $$ = ast_add_node_string(AST_SYSTEM, NULL, $3);
              free(yylval.string); }
           ;

relop : EQ { $$ = ast_add_node_leaf(AST_EQ); }
      | NE { $$ = ast_add_node_leaf(AST_NE); }
      | LT { $$ = ast_add_node_leaf(AST_LT); }
      | LE { $$ = ast_add_node_leaf(AST_LE); }
      | GT { $$ = ast_add_node_leaf(AST_GT); }
      | GE { $$ = ast_add_node_leaf(AST_GE); }
      ;

%%

void yyerror(char *s)
{
  /* Not the most effective method, but it helps alot. */
  if (parse_line < 0) {
    fprintf(stderr, "%s: error at basic input: %s\n",
      __FILE__, s);
  } else {
    fprintf(stderr, "%s: error after basic line %d: %s\n",
      __FILE__, parse_line, s);
  }
}

